from typing import Any, Dict
from os import environ as os_environ
from .functions import get_resource_content
from .arguments import ArgparseFormatter

SCROLLING_SPEED: int = 300
SCROLLING_ACCELERATION_DISTANCE: int = 10
SCROLLING_SLEEP_INTERVAL_INITIAL: float = 0.1
SCROLLING_DEAD_AREA: int = 50

BUTTONS_START: int = 2
BUTTONS_HOLD: bool = False

ICON_ENABLE: bool = False
ICON_SIZE: int = 30
ICON_PATH: str = 'resources/img/icon.svg'
ICON_ERROR: str = ('icon is enabled (it is disabled by default), but the '
                   '\'pyside6\' package is not installed. '
                   'remove \'--icon-enable\' or install the package\n'
                   '\npip install pyside6\n')

CONFIG_PATH: str = f'{os_environ.get("HOME")}/.config/autoscroll/config.txt'
CONFIG_ENABLE: bool = False
CONFIG_INTERVAL: int = 5
CONFIG_ERROR_ENABLE: str = 'you are trying to enable the config (\'enable\' is set to \'True\'), but the path is not valid'
CONFIG_ERROR_PARSE: str = 'you are trying to parse the config file, but \'enable\' is \'False\''


DEBUG_SCROLL: bool = False
DEBUG_CLICK: bool = False
DEBUG_INITIAL: bool = False
DEBUG_FILE: bool = False
DEBUG_PADDING: int = 16

COORDINATE_NAME: str = 'coordinates'


PARSER_INITIALIZER: Dict[str, Any] = {
    'prog': 'linux-xorg-autoscroll',
    'formatter_class': ArgparseFormatter,
    'description': get_resource_content('resources/txt/prolog.txt'),
    'fromfile_prefix_chars': '@'}

ARGUMENTS: Dict[str, Any] = {
    'scrolling': {
        'speed': {
            'type': int,
            'help': ('R|constant part of the scrolling speed\n'
                     f'[default: {SCROLLING_SPEED}]'),
        },
        'dead-area': {
            'type': int,
            'help': ('R|size of the square area aroung the starting '
                     'point where scrolling will stop, in pixels\n'
                     f'[default: {SCROLLING_DEAD_AREA}]')
        },
        'acceleration': {
            'type': int,
            'help': ('R|dynamic part of the  scrolling speed, depends on the '
                     'distance from the point where the scrolling started, '
                     'can be set to 0\n'
                     f'[default: {SCROLLING_ACCELERATION_DISTANCE}]')
        }
    },
    'buttons': {
        'hold': {
            'action': 'store_const',
            'const': True,
            'help': ('R|if set, the scrolling will end once you release '
                     '--buttons-start')
        },
        'start': {
            'type': int,
            'help': f'R|button that starts the scrolling\n[default: {BUTTONS_START}]'
        },
        'end': {
            'type': int,
            'help': ('R|button that ends the scrolling\n'
                     '[default: --buttons-start]')
        }
    },
    'config': {
        'enable': {
            'action': 'store_const',
            'const': True,
            'help': ('R|if set, arguments from the configuration file '
                     'on --config-path will be loaded every --config-interval')
        },
        'path': {
            'type': str,
            'help': f'R|path to the configuration file\n[default: {CONFIG_PATH}]',
        },
        'interval': {
            'type': int,
            'help': ('R|how often the config file should be checked for '
                     f'changes, in seconds\n[default: {CONFIG_INTERVAL}]')
        }
    },
    'icon': {
        'enable': {
            'action': 'store_const',
            'const': True,
            'help': 'if set, the icon will be enabled'
        },
        'path': {
            'type': str,
            'help': f'R|path to the icon\n[default: {ICON_PATH}]'
        },
        'size': {
            'type': int,
            'help': f'R|size of the icon, in pixels\n[default: {ICON_SIZE}]'
        }
    },
    'debug': {
        'file': {
            'action': 'store_const',
            'const': True,
            'help': ('if set, every time the config file is parsed, '
                     'information will be printed to stdout')
        },
        'click': {
            'action': 'store_const',
            'const': True,
            'help': 'if set, click info will be printed to stdout'
        },
        'scroll': {
            'action': 'store_const',
            'const': True,
            'help': 'if set, scroll info will be printed to stdout'
        },
        'initial': {
            'action': 'store_const',
            'const': True,
            'help': 'if set, startup configuration will be printed to stdout'
        }
    }
}



from argparse import SUPPRESS, HelpFormatter, ArgumentParser, _ArgumentGroup
from typing import Any, Dict, List


def parse_arguments(**arguments: Any) -> Dict[str, Any]:
    result = {}
    for key, value in arguments.items():
        key_split = key.split('_')
        group = key_split[0]
        name = '_'.join(key_split[1:])
        if group not in result:
            result[group] = {}
        result[group][name] = value
    return result


class ArgparseFormatter(HelpFormatter):

    # do not split lines that start with 'R|'
    # https://stackoverflow.com/questions/3853722/how-to-insert-newlines-on-argparse-help-text
    def _split_lines(self, text: str, width: int):
        if not text.startswith('R|'):
            return super()._split_lines(text, width)
        result = []
        for line in text[2:].splitlines(keepends=True):
            result.extend(super()._split_lines(line, width))
        return result

    # do not format the description
    # built-in argparse class argparse.RawDescriptionHelpFormatter
    def _fill_text(self, text: str, width: int, indent: str):
        return '\n'.join(indent + line
                         for line in text.splitlines(keepends=True))

    # change how 'metavar' is displayed
    # https://stackoverflow.com/questions/23936145/python-argparse-help-message-disable-metavar-for-short-options
    def _format_action_invocation(self, action):
        if not action.option_strings:
            return self._metavar_formatter(action, action.dest)(1)[0]
        # option takes no arguments -> -s, --long
        # option takes arguments:
        #    default output -> -s ARGS, --long ARGS
        #    changed output -> -s, --long type
        return ', '.join(action.option_strings) + (f' {action.type.__name__}'
                                                   if action.nargs != 0 else '')

    # add default value to the end
    # built-in argparse class argparse.ArgumentDefaultsHelpFormatter
    # def _get_help_string(self, action):
    #    if action.nargs == 0 or action.default is SUPPRESS or not action.default:
    #        return action.help
    #    return f'{action.help}\n[default: %(default)s]'


class _ArgparseArgumentGroup(_ArgumentGroup):

    def add_arguments(self,
                      **arguments: Dict[str, Any]) -> '_ArgparseArgumentGroup':
        group = self.title.split()[0].lower()
        for name, kwargs in arguments.items():
            flags = f'-{group[0]}{name[0]}', f'--{group}-{name}'
            self.add_argument(*flags, **kwargs)
        return self


class ArgparseParser(ArgumentParser):

    # override
    def add_argument_group(self, *args,
                           parameters: Dict[str, Dict[str, Any]] = None,
                           **kwargs):
        group = _ArgparseArgumentGroup(self, *args, **kwargs)
        self._action_groups.append(group)
        if parameters is not None:
            group.add_arguments(**parameters)
        return group

    # add a bunch of arguments and argument groups in one go
    def add_arguments(self, **groups: Dict[str, Any]) -> 'ArgparseParser':
        for name, parameters in groups.items():
            self.add_argument_group(title=name,
                                    description='',
                                    parameters=parameters)
        return self

    # change how '@' files are parsed
    # allows to have several arguments on one line
    # https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser.convert_arg_line_to_args
    def convert_arg_line_to_args(self, arg_line: str) -> List[str]:
        return arg_line.split()



#!/usr/bin/env python3

from sys import exit as sys_exit
from .autoscroll import Autoscroll


def start() -> None: sys_exit(Autoscroll().start(parse_argv=True))


if __name__ == '__main__':
    start()



from collections import ChainMap
from importlib.resources import path as importlib_path
from os.path import isfile as os_isfile
from typing import Any, Dict, Iterable, List, Tuple, Union


def construct_argument(argument: Tuple[str, Dict[str, str]]):
    help = '\n'.join([f'{key} ({value[0]}, default - {value[1]})' for key,
                      value in argument[1].items()])
    return ((f'-{argument[0][0]}', f'--{argument[0]}'),
            {'help': f'R|{help}', 'nargs': 2, 'action': 'append'})


def convert_bool(value: Union[str, bool] = None) -> bool:
    return value if isinstance(value, bool) else value.lower() in ('true', '+')


def return_none(*args, **kwargs) -> None: return


def return_kwargs(**kwargs) -> Dict[Any, Any]: return kwargs


def has_dict(*values: Any) -> bool:
    return False in (hasattr(value, '__dict__') for value in values)


def convert(value: Any) -> Dict[str, Any]:
    if value is None:
        return {}
    if isinstance(value, Iterable) and not isinstance(value, str):
        return value
    return raise_type_error(value, (type(None), Iterable))


def construct(item: Union[str, Tuple[str, Any]]) -> Dict[str, Any]:
    return {item: True} if isinstance(item, str) else {item[0]: item[1]}


def parse_argument(argument: List[Tuple[str, Any]]) -> Dict[str, Any]:
    return dict(ChainMap(*[construct(item) for item in convert(argument)]))


def parse_arguments_old(arguments: Dict[str, Any]) -> Dict[str, Any]:
    return {name: parse_argument(value) for name, value in arguments.items()}


def construct_coordinates(x: int, y: int) -> str: return f'{{{x}, {y}}}'


def documented_by(original):
    def wrapper(target):
        target.__doc__ = original.__doc__
        return target
    return wrapper


def raise_type_error(value: Any, target_type: Any) -> Any:
    if isinstance(value, target_type):
        return value
    message = f'value {str(value)} is not {target_type}, it is {type(value)}'
    raise TypeError(message)


def check_iterable(value: Iterable, length: int = 2) -> Iterable:
    raise_type_error(value, Iterable)
    if len(value) >= length:
        return value
    message = f'value "{str(value)}" does not have the length of {str(length)}'
    raise ValueError(message)


def get_path(path: str) -> str:
    return path if os_isfile(path) else get_resource_path(path)


def get_resource_path(resource: str = None) -> str:
    if resource is None:
        file = ''
        addition = ''
    else:
        split = resource.split('/')
        addition = '.'.join(split[0:-1])
        addition = f'.{addition}' if addition else ''
        file = split[-1]
        file = file if file != addition else ''
    return str(importlib_path(f'autoscroll{addition}', file))


def get_resource_content(resource: str) -> str:
    with open(get_resource_path(resource), 'r') as file:
        return file.read()



from .support import Base, Config, Scrolling, Icon, Buttons, Debug
from typing import Any, Dict
from pynput.mouse import Button, Listener
from threading import Event, Thread
from time import sleep


class Autoscroll(Base):

    def __init__(self, *args, **kwargs) -> None:
        self.scrolling: Scrolling = Scrolling()
        self.icon: Icon = Icon()
        self.config: Config = Config()
        self.buttons: Buttons = Buttons()
        self.debug: Debug = Debug()
        self.event_end: Event = Event()

        # initial update
        self.update(*args, **kwargs)

        # threads
        # listen for mouse actions and update information accordingly
        self.thread_scroll_listener = Listener(on_move=self._on_move,
                                               on_click=self._on_click,
                                               daemon=True)
        # scroll
        self.thread_scroll_action = Thread(target=self._loop,
                                           args=(self._is_not_end,
                                                 self._scroll),
                                           daemon=True)
        # listen for changes in the config file (if enabled)
        self.thread_config = Thread(target=self._loop,
                                    args=(self._is_not_end,
                                          self._update_from_config_file),
                                    daemon=True)

    def start(self, parse_argv: bool = False) -> None:
        # update from command line
        # it is a thread because in order to create an icon widget,
        # a qt application has has to be running
        # (qt applications have to run in the main thread)
        Thread(target=self.update,
               kwargs=self.config.parse_argv() if parse_argv else {}).start()
        # start listening for mouse movements and clicks
        self.thread_scroll_listener.start()
        # start the scrolling loop
        self.thread_scroll_action.start()
        # update from the command line
        # start listening for changes in the config file (if enabled)
        if self.config.enable:
            self.thread_config.start()
        # debug
        self._print('initial', self.debug.initial)
        # if the icon is not enabled, the main thread just waits
        self.icon.start_qt_when_icon_is_enabled()

    def _update_from_config_file(self) -> None:
        # update from the config file
        self.update(**self.config.parse_config_file())
        # debug
        self.config._print('config', self.debug.file, ['content'])
        # sleep
        sleep(self.config.interval)

    def _scroll(self) -> None:
        # wait for the scrolling event to be set in _on_click
        self.scrolling.wait()
        # wait for a period of time, calculated in _on_move
        self.scrolling.sleep_for_interval()
        # scroll on x-axis and y-axis, each scroll is either 1px, 0px, or -1px
        self.scrolling.scroll_once()

    def _on_move(self, x: int, y: int) -> None:
        # update coordinates
        self.scrolling.set_direction_and_coordinates(x, y)
        # recalculate sleep interval
        self.scrolling.set_interval()

    def _on_click(self, x: int, y: int, button: Button, pressed: bool) -> None:
        # send information about which button was pressed/released
        self.buttons.press(button, pressed)

        if (not self.scrolling.is_scrolling()
                and self.buttons.was_start_pressed()):
            self.scrolling.set_initial_coordinates(x, y)
            self.icon.show(x, y)
            self.scrolling.start()
        elif (self.buttons.was_end_pressed()
              or self.buttons.was_start_released_with_hold()):
            self.scrolling.stop()
            self.icon.hide()

        # it should be placed at the end to avoid initial scroll jumps
        self.scrolling.set_direction_and_coordinates(x, y)
        # debug
        self._print('click', self.debug.click)
        # start and end event have ended
        self.scrolling.clear_started_and_ended()
        # clear press information
        self.buttons.press_clear()

    def update(self,
               scrolling: Dict[str, Any] = None,
               icon: Dict[str, Any] = None,
               buttons: Dict[str, Any] = None,
               debug: Dict[str, Any] = None,
               config: Dict[str, Any] = None) -> None:
        self.config.update(**self._convert(config, {}, dict))
        self.scrolling.update(**self._convert(scrolling, {}, dict))
        self.icon.update(**self._convert(icon, {}, dict))
        self.buttons.update(**self._convert(buttons, {}, dict))
        self.debug.update(**self._convert(debug, {}, dict))

    def _is_not_end(self) -> bool: return not self.event_end.is_set()

    def json(self) -> Dict[str, Any]:
        return {'scrolling': self.scrolling, 'buttons': self.buttons,
                'icon': self.icon, 'debug': self.debug, 'config': self.config}



from PySide6.QtSvgWidgets import QSvgWidget
from PySide6.QtWidgets import QApplication
from PySide6.QtCore import Qt
from sys import argv as sys_argv
from .functions import get_path

application = QApplication(sys_argv)
application.setQuitOnLastWindowClosed(False)


class Icon(QSvgWidget):

    def __init__(self, path: str, size: int) -> None:
        super().__init__()
        self.update_icon(path, size)
        self.renderer().setAspectRatioMode(Qt.KeepAspectRatio)
        self.setWindowFlags(Qt.FramelessWindowHint
                            | Qt.WindowStaysOnTopHint
                            | Qt.WindowTransparentForInput
                            | Qt.Tool)
        self.setStyleSheet("background:transparent")
        self.setAttribute(Qt.WA_TranslucentBackground)

    def update_icon(self, path: str, size: int) -> None:
        self.load(get_path(path))
        self.icon_size = size
        self.resize(size, size)

    def show(self, x_current: int, y_current: int) -> None:
        half_size = self.icon_size // 2
        self.move(x_current - half_size, y_current - half_size)
        super().show()



#!/usr/bin env python3
# https://setuptools.pypa.io/en/latest/userguide/quickstart.html

from pathlib import Path
from setuptools import setup, find_packages

# The directory containing this file
HERE = Path(__file__).parent

# The text of the README file
README = (HERE / 'README.md').read_text()

# This call to setup() does all the work
setup(
    name='autoscroll',
    version='1.0.0',
    description='Enables universal autoscroll',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/kongrentian/autoscroll',
    author='kongrentian',
    author_email='causelesscause@gmail.com',
    license='GNU General Public License v3.0',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent'],
    packages=find_packages(where='.'),
    include_package_data=True,
    install_requires=['pynput'],
    extras_require={'icon': ['pyside6']},
    entry_points={'console_scripts': ['autoscroll = autoscroll.main:start']},
    python_requires='>=3.6')



from sys import argv as sys_argv
from threading import Event
from time import sleep
from pynput.mouse import Button, Controller, Listener
from .constants import (
    ARGUMENTS,
    BUTTONS_HOLD,
    BUTTONS_START,
    CONFIG_ENABLE,
    CONFIG_ERROR_ENABLE,
    CONFIG_PATH,
    CONFIG_INTERVAL,
    COORDINATE_NAME,
    DEBUG_CLICK,
    DEBUG_FILE,
    DEBUG_INITIAL,
    DEBUG_PADDING,
    DEBUG_SCROLL,
    ICON_ENABLE,
    ICON_ERROR,
    ICON_PATH,
    ICON_SIZE,
    PARSER_INITIALIZER,
    SCROLLING_ACCELERATION_DISTANCE,
    SCROLLING_DEAD_AREA,
    SCROLLING_SLEEP_INTERVAL_INITIAL,
    SCROLLING_SPEED)
from .functions import check_iterable, construct_coordinates, convert_bool, has_dict, raise_type_error, return_kwargs, return_none
from .arguments import ArgparseParser, parse_arguments
from typing import Any, Callable, Dict, Iterable, List, Tuple, Type, Union
from threading import Event, Thread
from os import stat as os_stat


class Base:

    debug_keys_ignore = []
    debug_keys_only = []

    def update(self, *args, **kwargs) -> None: ...

    def json(self) -> Dict[str, Any]: ...

    def __init__(self, *args, **kwargs) -> None: self.update(*args, **kwargs)

    def _set_if_nonexistent(self, name: str, value: Any) -> None:
        return None if hasattr(self, name) else setattr(self, name, value)

    @staticmethod
    def _convert_callable(value: Any) -> Any: return value

    def _set(self, _name: str, _nonexistent_value: Any, _value: Any,
             _type: Any = type(None), _callable: Callable = _convert_callable,
             **kwargs) -> None:
        self._set_if_nonexistent(_name, _nonexistent_value)
        value = self._convert(_value, getattr(self, _name), _type,
                              _callable, **kwargs)
        setattr(self, _name, value)

    def _convert(self, _value: Any, none_value: Any = None,
                 _type: Any = type(None),
                 _callable: Callable = _convert_callable, **kwargs) -> Any:
        """
        converts

        value is None: default

        target_type is None or value is target_type: value

        value is convert_type and convert function is callable:
            convert_function(value)

        raise_type_error
        """
        if isinstance(_value, _type) and callable(_callable):
            return _callable(_value, **kwargs)
        if _value is None:
            return none_value
        if isinstance(_value, _type):
            raise ValueError(f'value "{_value}" is "{_type}", but the '
                             'conversion function is not callable')
        return raise_type_error(_value, (type(None), _type))

    def _loop(self, condition: Callable = return_none,
              action: Callable = return_none,
              condition_getter: Callable = return_kwargs,
              action_getter: Callable = return_kwargs,
              condition_parameters: Dict[str, Any] = {},
              action_parameters: Dict[str, Any] = {}) -> None:

        if False in list(map(callable, (condition, action, condition_getter,
                                        action_getter))):
            raise TypeError(f'some functions are not callable')
        if not has_dict(condition_parameters, action_parameters):
            raise TypeError('parameters should have \'__dict__\' attribute')
        while condition(**condition_getter(**condition_parameters)):
            action(**action_getter(**action_parameters))

    def _print(self, header: str, do_print: bool = True,
               keys_only: List[str] = None,
               keys_ignore: List[str] = None) -> str:
        result = f'\n[{header}]\n{self._debug(keys_only, keys_ignore)}'
        if do_print:
            print(result)
        return result

    def __repr__(self) -> str: return self._debug()

    def __str__(self) -> str: return self._debug()

    def _debug(self, keys_only: List[str] = None,
               keys_ignore: List[str] = None) -> str:
        name = self.name if hasattr(self, 'name') else type(self).__name__
        raise_type_error(name, str)
        debug = []
        not_base = {}
        for key, value in self.json().items():
            if not self._debug_key_is_valid(key, keys_only, keys_ignore):
                continue
            if isinstance(value, Base):
                debug.append(str(value))
                continue
            not_base[key] = value if isinstance(value, str) else str(value)
        if not_base:
            debug.append(self._construct_debug(name.capitalize(), **not_base))
        return '\n'.join(debug)

    def _debug_key_is_valid(self, name: str, keys_only: List[str] = None,
                            keys_ignore: List[str] = None) -> bool:
        only = self.debug_keys_only if keys_only is None else keys_only
        ignore = self.debug_keys_ignore if keys_ignore is None else keys_ignore
        return name in only if only else name not in ignore

    def _construct_debug(self, _header: str, *args, **kwargs) -> str:
        result = ', '.join([item for item in args] +
                           [f'{name} - {value}'
                           for name, value in kwargs.items()])
        return _header.ljust(DEBUG_PADDING, ' ') + result


class Coordinate(Base):

    def update(self, current: Union[int, str] = None,
               previous: Union[int, str] = None,
               initial: Union[int, str] = None) -> None:
        self.initial: int = initial
        self.previous: int = previous
        self.current: int = current

    def distance(self, absolute: bool = False) -> int:
        distance = self.initial - self.current
        return abs(distance) if absolute else distance

    def direction(self) -> int:
        # direction > 0 -> 1, direction == 0 -> 0, direction < 0 -> -1
        return (self.distance() == 0) + (1 if self.distance() > 0 else -1)

    @property
    def current(self) -> int: return self._current

    @property
    def previous(self) -> int: return self._previous

    @property
    def initial(self) -> int: return self._initial

    @current.setter
    def current(self, value: int) -> None:
        _current = getattr(self, '_current', None)
        self._set('_current', 0, value, (str, int), int)
        self._set('_previous', 0, _current, (str, int), int)

    @previous.setter
    def previous(self, value: int) -> None:
        self._set('_previous', 0, value, (str, int), int)

    @initial.setter
    def initial(self, value: int) -> None:
        self._set('_initial', 0, value, (str, int), int)

    def json(self) -> Dict[str, Any]:
        return {'current': self.current, 'previous': self.previous,
                'initial': self.initial}


class Coordinates(Base):

    def update(self, x: Union[Coordinate, str, int] = None,
               y: Union[Coordinate, str, int] = None,
               name: str = None) -> None:
        self.x: Coordinate = x
        self.y: Coordinate = y
        self.name: str = self._convert(name, COORDINATE_NAME, str)

    def direction(self) -> Tuple[int, int]:
        return (self.x.direction(), self.y.direction())

    def distance(self, absolute: bool = False,
                 reverse: bool = False) -> Tuple[int, int]:
        x, y = self.x.distance(absolute), self.y.distance(absolute)
        return (-x, -y) if reverse else (x, y)

    @property
    def x(self) -> Coordinate: return self._x

    @property
    def y(self) -> Coordinate: return self._y

    @property
    def current(self) -> Tuple[int, int]:
        return (self.x.current, self.y.current)

    @property
    def initial(self) -> Tuple[int, int]:
        return (self.x.initial, self.y.initial)

    @property
    def previous(self) -> Tuple[int, int]:
        return (self.x.previous, self.y.previous)

    @y.setter
    def y(self, value: Union[str, int, Coordinate]) -> None:
        self._set('_y', Coordinate(), value, (Coordinate, str, int),
                  self._convert_coordinate, name='_y')

    @x.setter
    def x(self, value: Union[str, int, Coordinate]) -> None:
        self._set('_x', Coordinate(), value, (Coordinate, str, int),
                  self._convert_coordinate, name='_x')

    @current.setter
    def current(self, value: Tuple[Union[str, int], Union[str, int]]) -> None:
        self.x.current, self.y.current = self._convert(
            value, (None, None), Iterable, self._convert_iterable)

    @initial.setter
    def initial(self, value: Tuple[Union[str, int], Union[str, int]]) -> None:
        self.x.initial, self.y.initial = self._convert(
            value, (None, None), Iterable, self._convert_iterable)

    @previous.setter
    def previous(self, value: Tuple[Union[str, int], Union[str, int]]) -> None:
        self.x.previous, self.y.previous = self._convert(
            value, (None, None), Iterable, self._convert_iterable)

    def _convert_coordinate(self, value: Union[Coordinate, int],
                            name: str) -> Coordinate:
        if isinstance(value, (int, str)):
            coordinate = getattr(self, name)
            setattr(coordinate, 'current', int(value))
            return coordinate
        return raise_type_error(value, Coordinate)

    def _convert_iterable(self, value: Any) -> Tuple[int, int]:
        raise_type_error(value, Iterable)
        x, y = None, None
        if len(value) >= 1:
            x = self._convert(value[0], None, (str, int), int)
        if len(value) >= 2:
            y = self._convert(value[1], None, (str, int), int)
        return x, y

    def json(self) -> Dict[str, Any]:
        coordinates = {'direction': str(self.direction()),
                       'distance': str(self.distance())}
        coordinates.update({item: construct_coordinates(getattr(self.x, item),
                                                        getattr(self.y, item))
                            for item in ('current', 'previous', 'initial')})
        return coordinates


class Buttons(Base):

    def __init__(self, *args, **kwargs) -> None:
        self.update(*args, **kwargs)

        self.button: Button = None
        self.is_pressed: bool = None

    def update(self, start: Union[Button, int, str] = None,
               end: Union[Button, int, str] = None,
               hold: Union[bool, str] = None) -> None:
        self.start: Button = start
        self.end: Button = end
        self.hold: bool = hold

    def press(self, button: Button, pressed: bool) -> None:
        self.button, self.is_pressed = button, pressed

    def press_clear(self) -> None: self.button, self.is_pressed = None, None

    def is_start(self) -> bool: return self.start == self.button

    def is_end(self) -> bool: return self.end == self.button

    def was_action(self) -> bool:
        return self.button is not None and self.is_pressed is not None

    def was_start_pressed(self) -> bool:
        return self.was_action() and self.is_start() and self.is_pressed

    def was_end_pressed(self) -> bool:
        return self.was_action() and self.is_end() and self.is_pressed

    def was_start_released(self) -> bool:
        return self.was_action() and self.is_start() and not self.is_pressed

    def was_start_released_with_hold(self) -> bool:
        return self.was_start_released() and self.hold

    def was_end_released(self) -> bool:
        return self.was_action() and self.is_end() and not self.is_pressed

    @property
    def start(self) -> Button: return self._start

    @property
    def hold(self) -> bool: return self._hold

    @property
    def end(self) -> Button: return self._end

    @start.setter
    def start(self, value: Union[int, str, Button] = None) -> None:
        self._set('_start', Button(BUTTONS_START), value, (int, Button, str),
                  self._convert_button)

    @end.setter
    def end(self, value: Union[int, str, Button] = None) -> None:
        self._set('_end', self.start, value, (int, Button, str),
                  self._convert_button)

    @hold.setter
    def hold(self, value: Union[str, bool] = None) -> None:
        self._set('_hold', BUTTONS_HOLD, value, (str, bool), convert_bool)

    @staticmethod
    def _convert_button(value: Union[Button, int, str]) -> Button:
        return value if isinstance(value, Button) else Button(int(value))

    def json(self) -> Dict[str, Any]:
        return {'start': self.start.name, 'end': self.end.name,
                'hold': self.hold, 'pressed button': self.button,
                'pressed': self.is_pressed}


class Scrolling(Base):

    def __init__(self, *args, **kwargs) -> None:
        self.sleep_interval: int = SCROLLING_SLEEP_INTERVAL_INITIAL
        self.controller: Controller = Controller()
        self.event_end: Event = Event()
        self.event_scrolling: Event = Event()
        self.event_started: Event = Event()
        self.event_ended: Event = Event()

        self.coordinates: Coordinates = Coordinates()
        self.coordinates.debug_keys_ignore = ['direction']
        self.direction: Coordinates = Coordinates(name='direction')
        self.direction.debug_keys_only = ['direction']

        self.update(*args, **kwargs)

    def update(self, dead_area: Union[str, int] = None,
               speed: Union[str, int] = None,
               acceleration: Union[str, int] = None) -> None:
        self.speed: int = speed
        self.dead_area: int = dead_area
        self.acceleration: int = acceleration

    def sleep_for_interval(self) -> None: sleep(self.sleep_interval)

    def wait(self) -> None: self.event_scrolling.wait()

    def scroll_once(self) -> None:
        self.controller.scroll(*self.direction.current)

    def start(self) -> None:
        self.event_started.set()
        self.event_scrolling.set()

    def stop(self) -> None:
        self.event_scrolling.clear()
        self.event_ended.set()

    def clear_started_and_ended(self) -> None:
        self.event_ended.clear()
        self.event_started.clear()

    def is_scrolling(self) -> bool: return self.event_scrolling.is_set()

    def has_started(self) -> bool: return self.event_started.is_set()

    def has_ended(self) -> bool: return self.event_ended.is_set()

    def is_not_end(self) -> bool: return not self.event_end.is_set()

    def is_dead_area(self) -> bool:
        distance = self.coordinates.distance(absolute=True)
        return distance[0] <= self.dead_area and distance[1] <= self.dead_area

    def set_interval(self) -> None:
        distance = self.coordinates.distance(absolute=True)
        interval = self.acceleration * max(distance) + self.speed
        self.sleep_interval = abs(100 / interval) \
            if interval else SCROLLING_SLEEP_INTERVAL_INITIAL

    def set_initial_coordinates(self, x: int, y: int) -> None:
        self.coordinates.initial = x, y

    def set_direction_and_coordinates(self, x: int, y: int) -> None:
        self.coordinates.update(x, y)
        self.direction.current = ((0, 0) if self.is_dead_area()
                                  else self.coordinates.direction())

    def json(self) -> str:
        return {'active': self.is_scrolling(),
                'interval': self.sleep_interval,
                'acceleration': self.acceleration,
                'dead_area': self.dead_area,
                'started': self.has_started(),
                'ended': self.has_ended(),
                'coordinates': self.coordinates,
                'direction': self.direction}

    @property
    def speed(self) -> int: return self._speed

    @property
    def dead_area(self) -> int: return self._dead_area

    @property
    def acceleration(self) -> int: return self._acceleration

    @speed.setter
    def speed(self, value: Union[str, int] = None) -> None:
        self._set('_speed', SCROLLING_SPEED, value, (str, int), int)

    @dead_area.setter
    def dead_area(self, value: Union[str, int]) -> None:
        self._set('_dead_area', SCROLLING_DEAD_AREA, value, (str, int), int)

    @acceleration.setter
    def acceleration(self, value: Union[str, int]) -> None:
        self._set('_acceleration', SCROLLING_ACCELERATION_DISTANCE, value,
                  (str, int), int)


class Icon(Base):

    def __init__(self, *args, **kwargs) -> None:
        self.application = None
        self.event_icon_enabled: Event = Event()
        self.event_qt_application_started: Event = Event()
        self.update(*args, **kwargs)

    def update(self, enable: Union[str, bool] = None,
               path: str = None, size: Union[str, int] = None) -> None:
        self.path: str = path
        self.size: int = size
        self.enable: bool = enable
        self.icon: Union[None, object] = self.path, self.size

    def show(self, x: int, y: int) -> None:
        return self.icon.show(x, y) if self.enable else None

    def hide(self) -> None: return self.icon.hide() if self.enable else None

    def json(self) -> Dict[str, Any]:
        return {'enable': self.enable, 'path': self.path, 'size': self.size}

    def start_qt_when_icon_is_enabled(self) -> None:
        self.event_icon_enabled.wait()
        self.application = self._get_qt(True)
        self.event_qt_application_started.set()
        self.application.exec()

    @property
    def path(self) -> str: return self._path

    @property
    def size(self) -> int: return self._size

    @property
    def enable(self) -> bool: return self._enable

    @property
    def icon(self) -> Union[None, Any]: return self._icon

    @path.setter
    def path(self, value: str) -> None:
        self._set('_path', ICON_PATH, value, str)

    @size.setter
    def size(self, value: Union[str, int]) -> None:
        self._set('_size', ICON_SIZE, value, (str, int), int)

    @enable.setter
    def enable(self, value: Union[str, bool]) -> None:
        self._set('_enable', ICON_ENABLE, value, (str, bool), convert_bool)

    @icon.setter
    def icon(self, value: Tuple[str, int]) -> None:
        if not self.enable:
            self._icon = None
            return
        self.event_icon_enabled.set()
        self.event_qt_application_started.wait()
        value = check_iterable(value)
        if hasattr(self, '_icon') and self.icon is not None:
            self.icon.update_icon(*value)
            return
        self._icon = self._get_qt()(*value)
        self.application.setActiveWindow(self.icon)

    def _get_qt(self, get_application: bool = False) -> object:
        try:
            from .qt import Icon as qt_icon, application
        except ImportError as exception:
            raise ValueError(ICON_ERROR) from exception
        return application if get_application else qt_icon


class Debug(Base):

    def update(self, scroll: bool = None, file: bool = None,
               click: bool = None, initial: bool = None) -> None:
        self.scroll: bool = scroll
        self.click: bool = click
        self.initial: bool = initial
        self.file: bool = file

    def json(self) -> Dict[str, Any]:
        return {'scroll': self.scroll, 'click': self.click,
                'initial': self.initial, 'file': self.file}

    @property
    def scroll(self) -> bool: return self._scroll

    @property
    def click(self) -> bool: return self._click

    @property
    def initial(self) -> bool: return self._initial

    @property
    def file(self) -> bool: return self._file

    @scroll.setter
    def scroll(self, value: bool) -> None:
        self._set('_scroll', DEBUG_SCROLL, value, (str, bool), convert_bool)

    @click.setter
    def click(self, value: bool) -> None:
        self._set('_click', DEBUG_CLICK, value, (str, bool), convert_bool)

    @initial.setter
    def initial(self, value: bool) -> None:
        self._set('_initial', DEBUG_INITIAL, value, (str, bool), convert_bool)

    @file.setter
    def file(self, value: bool) -> None:
        self._set('_file', DEBUG_FILE, value, (str, bool), convert_bool)


class Config(Base):

    debug_keys_ignore = 'content'

    def __init__(self, *args, **kwargs) -> None:
        self._stamp: int = 0
        self._parse_config_file_content: Dict[str, Any] = {}
        self.argument_parser: ArgparseParser = ArgparseParser(
            **PARSER_INITIALIZER).add_arguments(**ARGUMENTS)
        self.update(*args, **kwargs)

    def update(self, enable: Union[bool, str] = None,
               path: str = None, interval: Union[str, int] = None) -> None:
        self.path: str = path
        self.enable: bool = enable
        self.interval: int = interval

    def parse_argv(self) -> Dict[str, Any]: return self._parse()

    def parse_string(self, value: str) -> Dict[str, Any]:
        return self._parse(value.split())

    def parse_config_file(self) -> Dict[str, Any]:
        self._parse_config_file_content = self._parse_config_file()
        return self._parse_config_file_content

    def _parse_config_file(self) -> Dict[str, Any]:
        if not self._has_file_changed():
            return {}
        with open(self.path, 'r') as config_file:
            config = config_file.read()
        result = self.parse_string(config.replace('\n', ' '))
        return result

    def _parse(self, *args, **kwargs) -> Dict[str, Any]:
        return parse_arguments(**vars(
            self.argument_parser.parse_args(*args, **kwargs)))

    def _has_file_changed(self) -> bool:
        stamp = os_stat(self.path).st_mtime
        if stamp == self._stamp:
            return False
        self._stamp = stamp
        return True

    @property
    def enable(self) -> bool: return self._enable

    @property
    def path(self) -> str: return self._path

    @property
    def interval(self) -> int: return self._interval

    @enable.setter
    def enable(self, value: Union[bool, str]) -> None:
        self._set('_enable', CONFIG_ENABLE, value, (bool, str), convert_bool)
        if self._enable and not self.path:
            self._enable = False
            raise ValueError(f'{CONFIG_ERROR_ENABLE}, path - {self.path}')

    @interval.setter
    def interval(self, value: Union[str, int]) -> None:
        self._set('_interval', CONFIG_INTERVAL, value, (int, str), int)

    @path.setter
    def path(self, value: str) -> None:
        self._set('_path', CONFIG_PATH, value, str)

    def json(self) -> Dict[str, Any]:
        return {'path': self.path, 'enable': self.enable,
                'interval': self.interval,
                'content': self._parse_config_file_content}
